import dask.dataframe.multi
from dask.dataframe import DataFrame
from pandas import DataFrame as PDataFrame
from typing import List, Optional
import distributed
from framania.pandasmania.merge import merge_on_columns_without_breaking_index


def merge_pandas_df_on_columns_without_breaking_index(left_df: DataFrame, right_df: PDataFrame,
                                                      on: Optional[List[str]] = None,
                                                      left_on: Optional[List[str]] = None,
                                                      right_on: Optional[List[str]] = None,
                                                      how: str = 'inner',
                                                      keep_left_index: bool = True,
                                                      keep_right_index: bool = True,
                                                      reindex_by_left: bool = True,
                                                      reindex_by_right: bool = False) \
        -> DataFrame:
    return left_df \
        .map_partitions(
        lambda df: merge_on_columns_without_breaking_index(df, right_df, on,
                                                           left_on, right_on,
                                                           how, keep_left_index, keep_right_index,
                                                           reindex_by_left, reindex_by_right))


def merge_pandas_df_future_on_columns_without_breaking_index(left_df: DataFrame, right_df: 'distributed.client.Future',
                                                             on: Optional[List[str]] = None,
                                                             left_on: Optional[List[str]] = None,
                                                             right_on: Optional[List[str]] = None,
                                                             how: str = 'inner',
                                                             keep_left_index: bool = True,
                                                             keep_right_index: bool = True,
                                                             reindex_by_left: bool = True,
                                                             reindex_by_right: bool = False) \
        -> DataFrame:
    return left_df \
        .map_partitions(
        lambda df: merge_on_columns_without_breaking_index(df, right_df.result(),
                                                           on, left_on, right_on,
                                                           how, keep_left_index, keep_right_index,
                                                           reindex_by_left, reindex_by_right))


def merge_dask_df_on_columns_without_breaking_index(left_df: DataFrame, right_df: DataFrame,
                                                    on: Optional[List[str]] = None,
                                                    left_on: Optional[List[str]] = None,
                                                    right_on: Optional[List[str]] = None,
                                                    how: str = 'inner',
                                                    keep_left_index: bool = True,
                                                    keep_right_index: bool = True,
                                                    reindex_by_left: bool = True,
                                                    reindex_by_right: bool = False):
    assert left_df.divisions == right_df.divisions
    return dask.dataframe.multi.map_partitions(
        lambda left, right: merge_on_columns_without_breaking_index(left, right,
                                                                    on, left_on, right_on,
                                                                    how, keep_left_index, keep_right_index,
                                                                    reindex_by_left, reindex_by_right),
        left_df, right_df
    )



import shutil
from base64 import b64encode
from hashlib import sha1
from pathlib import Path
from typing import Union, List, Any, Optional
from uuid import uuid4

import dask.dataframe
import pandas


def _concat_dimension_and_add_partition_key(df: pandas.DataFrame,
                                            partition_on: str, partition_value: Any,
                                            dimension_row: pandas.Series,
                                            merge_keys: Optional[List[str]] = None):
    if merge_keys is None:
        merge_keys = []
    df['__merge__'] = 1
    dimension_row = pandas.DataFrame([dimension_row])
    dimension_row['__merge__'] = 1
    merged = dimension_row.merge(df, on=merge_keys + ['__merge__'], how='inner')
    del merged['__merge__']
    merged[partition_on] = partition_value
    return merged.set_index(partition_on)


def _read_parquet_then_concat_dimension_and_add_partition_key(parquet_path: Path,
                                                              partition_on: str, partition_value: Any,
                                                              dimension_row: pandas.Series,
                                                              merge_keys: Optional[List[str]] = None):
    df = pandas.read_parquet(parquet_path)
    return _concat_dimension_and_add_partition_key(df, partition_on, partition_value, dimension_row, merge_keys)


def _create_distributed_dimension_data(src_data: pandas.DataFrame, dimension_data: pandas.DataFrame,
                                       tmp_dir: Path):
    assert tmp_dir.is_dir()
    idx_name = dimension_data.index.name
    files = {}
    meta = None
    for idx, row in dimension_data.iterrows():
        dimension_concat_data = _concat_dimension_and_add_partition_key(src_data, idx_name, idx, row)
        if meta is None:
            meta = dimension_concat_data.iloc[:0]
        uid = str(uuid4())
        dimension_concat_data.to_parquet(tmp_dir / f"{uid}.parquet")
        files[idx] = uid
    return files, meta


def duplicate_pandas_for_each_distribution_key(pandas_df: pandas.DataFrame, dask_df: dask.dataframe.DataFrame,
                                               dask_dimension_columns: List[str],
                                               temporary_parquet_root: Union[str, Path],
                                               dimension_merge_keys: Optional[List[str]] = None,
                                               delete_existing_temporary_directory: bool = True):
    assert (dask_df.index.map_partitions(lambda i: i.nunique()).compute() == 1).all(), \
        "Please reindex dask dataframe as mono-index distribution using `set_mono_index_via_disk`"
    if delete_existing_temporary_directory:
        shutil.rmtree(str(temporary_parquet_root), ignore_errors=True)
    temporary_parquet_root.mkdir(parents=True, exist_ok=True)
    dimension_df = dask_df[dask_dimension_columns].map_partitions(lambda pdf: pdf.head(1)).compute()

    assert len(dimension_df) > 0, "Dask df must not be empty."

    parquet_file = temporary_parquet_root / "pandas_df.parquet"
    pandas_df.to_parquet(parquet_file)

    keys = dimension_df.index
    name = f"framania-duplicate-pandas-for-each-distribution-key-{str(uuid4())}"
    procs = {}

    meta = None
    for i, (idx, row) in enumerate(dimension_df.iterrows()):
        procs[(name, i)] = (_read_parquet_then_concat_dimension_and_add_partition_key,
                            parquet_file,
                            dimension_df.index.name, idx, row, dimension_merge_keys)
        if i == 0:
            meta = _read_parquet_then_concat_dimension_and_add_partition_key(
                parquet_file,
                dimension_df.index.name,
                idx, row, dimension_merge_keys)

    divisions = keys + [keys[-1]]

    out_dir = temporary_parquet_root / "merge_ddf"
    dask.dataframe.DataFrame(procs, name, meta, divisions).to_parquet(out_dir)
    return dask.dataframe.read_parquet(out_dir)


def _read_parquet_files_then_concat(files: List[Union[str, Path]], partition_on: str, partition_value: Any,
                                    options: dict):
    concat_result = pandas.concat([pandas.read_parquet(f, **options) for f in files], axis=0)
    concat_result[partition_on] = partition_value
    return concat_result.set_index(partition_on)


def _read_dask_dataframe_from_partitioned_parquets(parquet_dir_root: Union[str, Path], partition_on: str, **options):
    p = Path(parquet_dir_root)
    normal_ddf = dask.dataframe.read_parquet(parquet_dir_root)
    keys = list(sorted(normal_ddf[partition_on].drop_duplicates().compute()))

    name = f"framania-read-from-partitioned-parquets-{str(uuid4())}"
    procs = {}

    divisions = keys + [keys[-1]]

    partition_dtype = normal_ddf._meta[partition_on].cat.categories.dtype
    normal_ddf._meta[partition_on] = normal_ddf._meta[partition_on].astype(partition_dtype)

    meta = normal_ddf._meta.set_index(partition_on)

    for i, key in enumerate(keys):
        files = list((p / f"{partition_on}={key}").glob("part.*.parquet"))
        procs[(name, i)] = (_read_parquet_files_then_concat, files, partition_on, key, options)

    return dask.dataframe.DataFrame(procs, name, meta, divisions)


def set_index_via_disk(df: dask.dataframe.DataFrame, column: str,
                       temporary_parquet_root: Union[str, Path],
                       delete_existing_temporary_directory: bool = True,
                       **options):
    if delete_existing_temporary_directory:
        shutil.rmtree(str(temporary_parquet_root), ignore_errors=True)
    df.to_parquet(temporary_parquet_root, partition_on=column, **options)
    idx_df = _read_dask_dataframe_from_partitioned_parquets(temporary_parquet_root,
                                                            column, **options)
    return idx_df


set_mono_index_via_disk = set_index_via_disk


def _pandas_row_to_hash(row) -> str:
    s1 = sha1()
    for v in row:
        s1.update(str(v).encode())
    return b64encode(s1.digest()).decode('utf8')


def set_hash_index_via_disk(df: dask.dataframe.DataFrame, columns: List[str],
                            temporary_parquet_root: Union[str, Path],
                            drop_existing_index: bool = True,
                            delete_existing_temporary_directory: bool = True,
                            **options):
    if delete_existing_temporary_directory:
        shutil.rmtree(str(temporary_parquet_root), ignore_errors=True)
    index_keys = df[columns].drop_duplicates().compute()
    index_name = f'HASH-{"-".join([str(v) for v in columns])}'
    index_keys[index_name] = index_keys.apply(_pandas_row_to_hash, axis=1)

    meta = pandas.DataFrame({c: df._meta[c] for c in df.columns})
    meta[index_name] = pandas.Series([], dtype='str')
    df_ = df.map_partitions(lambda df: df.reset_index(drop=drop_existing_index).merge(index_keys, on=columns),
                            meta=meta)

    if df.known_divisions:
        df_ = df_.clear_divisions()
        df_._meta.reset_index(drop=True, inplace=True)

    return set_index_via_disk(df_, index_name, temporary_parquet_root, **options)



from typing import List, Any, Union, Tuple

from pandas import DataFrame


def drop_duplicates_by_named_index_and_keys(pd: DataFrame, subset: List[Any], keep: Union[str, bool] = 'first'):
    """
        API to do drop_duplicates with target subset including named index.

        Args:
            df (dask.dataframe.DataFrame): target dask dataframe
            subset (List[Any]): drop_duplicates subset key columns that be used with named index
            keep (Union[str, bool]): see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html
        Returns:
            result dask dataframe
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'idx': [0, 1, 2, 0, 1, 2]}).set_index('idx')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                   a    b
            idx
            0      1    0
            1     10  100
            2    100  200
            0      1  300
            1      1  400
            2    100  500
            >>> result = drop_duplicates_by_named_index_and_keys(pd, ['a'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
                   a    b
            idx
            0      1    0
            1     10  100
            1      1  400
            2    100  200
            >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100],
            ...                        'b': range(0, 600, 100),
            ...                        'idx1': [0, 1, 2, 0, 1, 2],
            ...                        'idx2': [1, 1, 1, 1, 1, 1]}).set_index(['idx1', 'idx2'])
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                         a    b
            idx1 idx2
            0    1       1    0
            1    1      10  100
            2    1     100  200
            0    1       1  300
            1    1       1  400
            2    1     100  500
            >>> result = drop_duplicates_by_named_index_and_keys(pd, ['a'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
                         a    b
            idx1 idx2
            0    1       1    0
            1    1      10  100
                 1       1  400
            2    1     100  200
        """
    index_all_levels = [l for l in range(pd.index.nlevels)]
    return pd.groupby(level=index_all_levels, group_keys=False).apply(lambda pd: pd.drop_duplicates(subset=subset))


def index_level_value(df: DataFrame, level_idx: Union[int, str]):
    """
    >>> df = DataFrame({'a': [1,2,3], 'b': [2,3,4], 'c': [3,4,5], 'i1': [0,0,1], 'i2': [0, 1, 0]})
    >>> df.set_index(['i1', 'i2'], inplace=True)
    >>> df
    ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c
        i1 i2
        0  0   1  2  3
           1   2  3  4
        1  0   3  4  5
    >>> index_level_value(df, 0)
    Int64Index([0, 0, 1], dtype='int64', name='i1')
    >>> index_level_value(df, 'i1')
    Int64Index([0, 0, 1], dtype='int64', name='i1')
    """
    return df.index.get_level_values(level_idx)


def illoc(df: DataFrame, level_idx: Union[int, str], value: Any):
    """
    >>> df = DataFrame({'a': [1,2,3], 'b': [2,3,4], 'c': [3,4,5], 'i1': [0,0,1], 'i2': [0, 1, 0]})
    >>> df.set_index(['i1', 'i2'], inplace=True)
    >>> df
    ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c
        i1 i2
        0  0   1  2  3
           1   2  3  4
        1  0   3  4  5
    >>> illoc(df, 1, 0)
    ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c
        i1 i2
        0  0   1  2  3
        1  0   3  4  5
    >>> illoc(df, 'i2', 0)
    ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c
        i1 i2
        0  0   1  2  3
        1  0   3  4  5
    """
    return df[df.index.get_level_values(level_idx) == value]


class ILLoc:
    def __init__(self, df: DataFrame):
        self.df = df

    def __getitem__(self, item: Tuple[Union[int, str], Any]):
        assert isinstance(item, tuple)
        assert len(item) == 2
        assert isinstance(item[0], int) or isinstance(item[0], str)
        level_index = index_level_value(self.df, item[0])
        return level_index.to_frame().loc[item[1]]


il = index_level_value
DataFrame.il = lambda self, idx: il(self, idx)
DataFrame.illoc = lambda self: ILLoc(self)



from typing import List, Optional
from pandas import DataFrame, MultiIndex


def merge_on_columns_without_breaking_index(left_df: DataFrame, right_df: DataFrame,
                                            on: Optional[List[str]]=None,
                                            left_on: Optional[List[str]]=None, right_on: Optional[List[str]]=None,
                                            how: str = 'inner',
                                            keep_left_index: bool = True,
                                            keep_right_index: bool = True,
                                            reindex_by_left: bool = True,
                                            reindex_by_right: bool = False) \
        -> DataFrame:
    """
    >>> df1 = DataFrame({"A": [1,2,3,4,5,1,2,3,4,5],
    ...                  "B": [2,2,3,3,4,4,5,5,6,6],
    ...                  "C": ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']}).set_index('A')
    >>> df2 = DataFrame({"a": [1,2,3,4,5,1,2,3,4,5],
    ...                  "b": [2,2,3,3,4,4,5,5,6,6],
    ...                  "c": ['aa', 'ba', 'ca', 'da', 'ea', 'fa', 'ga', 'ha', 'ia', 'ja']}).set_index('a')
    >>> result = merge_on_columns_without_breaking_index(df1, df2, left_on=["A", "B"], right_on=["a", "b"], how="inner")
    >>> result
    ... # doctest: +NORMALIZE_WHITESPACE
           B  C  a  b   c
        A
        1  2  a  1  2  aa
        2  2  b  2  2  ba
        3  3  c  3  3  ca
        4  3  d  4  3  da
        5  4  e  5  4  ea
        1  4  f  1  4  fa
        2  5  g  2  5  ga
        3  5  h  3  5  ha
        4  6  i  4  6  ia
        5  6  j  5  6  ja
    >>> df1 = DataFrame({"A": [1,2,3,4,5,1,2,3,4,5],
    ...                  "B": [2,2,3,3,4,4,5,5,6,6],
    ...                  "C": ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']}).set_index(["A", "B"])
    >>> df2 = DataFrame({"a": [1,2,3,4,5,1,2,3,4,5],
    ...                  "b": [2,2,3,3,4,4,5,5,6,6],
    ...                  "c": ['aa', 'ba', 'ca', 'da', 'ea', 'fa', 'ga', 'ha', 'ia', 'ja']}).set_index(['a', 'b'])
    >>> result = merge_on_columns_without_breaking_index(df1, df2, left_on=["A", "B"], right_on=["a", "b"], how="inner")
    >>> result
    ... # doctest: +NORMALIZE_WHITESPACE
             C  a  b   c
        A B
        1 2  a  1  2  aa
        2 2  b  2  2  ba
        3 3  c  3  3  ca
        4 3  d  4  3  da
        5 4  e  5  4  ea
        1 4  f  1  4  fa
        2 5  g  2  5  ga
        3 5  h  3  5  ha
        4 6  i  4  6  ia
        5 6  j  5  6  ja
    >>> result = merge_on_columns_without_breaking_index(df1, df2, left_on=["A", "B"], right_on=["a", "b"], how="inner",
    ...                                                  reindex_by_left=False, reindex_by_right=True)
    >>> result
    ... # doctest: +NORMALIZE_WHITESPACE
             A  B  C   c
        a b
        1 2  1  2  a  aa
        2 2  2  2  b  ba
        3 3  3  3  c  ca
        4 3  4  3  d  da
        5 4  5  4  e  ea
        1 4  1  4  f  fa
        2 5  2  5  g  ga
        3 5  3  5  h  ha
        4 6  4  6  i  ia
        5 6  5  6  j  ja
    """

    if (on is None) and ((left_on is None) or (right_on is None)):
        raise ValueError("onを指定するか、left_onとright_onの両方を指定する必要があります。")
    if on is not None:
        left_on = on
        right_on = on

    if isinstance(left_df.index, MultiIndex):
        left_name = left_df.index.names
    else:
        left_name = left_df.index.name

    if isinstance(right_df.index, MultiIndex):
        right_name = right_df.index.names
    else:
        right_name = right_df.index.name

    assert reindex_by_left ^ reindex_by_right
    if keep_left_index and keep_right_index:
        assert left_name != right_name

    reindex_by = None

    if reindex_by_left:
        assert left_name is not None
        reindex_by = left_name
    elif reindex_by_right:
        assert right_name is not None
        reindex_by = right_name

    if keep_left_index:
        assert left_name is not None
        left_df = left_df.reset_index()

    if keep_right_index:
        assert right_name is not None
        right_df = right_df.reset_index()

    merge_result = left_df.merge(right_df, left_on=left_on, right_on=right_on, how=how)
    if reindex_by_left or reindex_by_right:
        merge_result = merge_result.set_index(reindex_by)
    return merge_result
